/**
 * Created by John on 7/5/2017.
 *
 * Common logic for manipulating ASN models
 */

var moment = require("moment");

function AsnHelper(models) {
    this.models = models;
    this.winston = require("winston");
    require("../config/logging")(this.winston);
    this.logger = this.winston.loggers.get("transaction");
}

/**
 * Determines if the addr object has any attributes populated
 *
 * @param {Object} addr
 * @returns {boolean} true if at least one field is populated.
 */
AsnHelper.prototype.addressExists = function(addr) {
    if (!addr) return false;
    return (addr.accountId || addr.name || addr.address || addr.city || addr.state || addr.zip);
};

/**
 * Builds the AsnAddress model for persistence
 *
 * @param {Object} addr Address from JSON
 * @param {F/T} type (F)rom or (T)o address
 * @param {int} parentId ASN instance id
 * @returns {AsnAddress} Populated AsnAddress model
 */
AsnHelper.prototype.buildAddresssModel = function (addr, type, parentId) {
    return this.models.AsnAddress.build({
        asnId: parentId,
        shipType: type,
        accountId: addr.accountId,
        name: addr.name,
        address: addr.address,
        city: addr.city,
        state: addr.state,
        zip: addr.zip
    });
};
/**
 * Saves a single ASN record (with its addresses) to the database vi promise
 *
 * @param {Object} msg JSON from Http
 */
AsnHelper.prototype.saveAsn = function(msg) {
    var self = this;

    return new Promise(function (resolve, reject) {
        var asn = msg;
        var shipFrom = null;
        if (asn.shipFrom) {
            shipFrom = JSON.parse(JSON.stringify(asn.shipFrom));
            delete asn.shipFrom;
        }
        var shipTo = null;
        if (asn.shipTo) {
            shipTo = JSON.parse(JSON.stringify(asn.shipTo));
            delete asn.shipTo;
        }
        self.models.ASN.create(asn).then(function (doc) {

            var promises = [];

            if (self.addressExists(shipTo)) {
                promises.push(new Promise(function (stResolve, stReject) {
                    var addr = self.buildAddresssModel(shipTo, "T", doc.id);
                    addr.save().then(function (addrImpl) {
                        self.logger.debug("Ship To saved for " + doc.asnId);
                        stResolve(addrImpl);
                    }).catch(function (tErr) {
                        stReject(tErr);
                    });
                }));
            }
            if (self.addressExists(shipFrom)) {
                promises.push(new Promise(function (sfResolve, sfReject) {
                    var addr = self.buildAddresssModel(shipFrom, "F", doc.id);
                    addr.save().then(function (addrImpl) {
                        self.logger.debug("Ship From saved for " + doc.asnId);
                        sfResolve(addrImpl);
                    }).catch(function (fErr) {
                        sfReject(fErr);
                    });
                }));
            }

            if (promises.length > 0) {
                Promise.all(promises).then(function (addresses) {
                    var out = JSON.parse(JSON.stringify(doc));
                    out['Addresses'] = addresses;
                    resolve(out);
                }).catch(function (addrErr) {
                    reject(addrErr);
                });
            } else {
                resolve(doc);
            }
        }).catch(function (err) {
            reject(err);
        });
    });
};
/**
 * Saves an array of ASN JSON objects to the database (via promises)
 *
 * @param {Array} asns Collection of ASN JSON object promises
 */
AsnHelper.prototype.saveAsns = function(asns) {
    var self = this;

    if (!Array.isArray(asns)) {
        throw new Error("Invalid input");
    }

    var promises = [];
    for (var i=0; i< asns.length; i++) {
        promises.push(self.saveAsn(asns[i]));
    }

    return promises;

};
/**
 * Finds a single ASN by its ID
 *
 * @param {Number} id Primary Key
 * @param {Function} cb callback(err, asn)
 */
AsnHelper.prototype.getAsnById = function(id, cb) {
    var self = this;
    self.models.ASN.findOne({
        where: { id: id },
        include: [{ model: self.models.AsnAddress, as: 'Addresses' }]
    }).then(function(asn) {
        if (cb) cb(null, asn);
        else return asn;
    }).catch(function(err){
        if (cb) cb(err, null);
        else throw err;
    });
};
/**
 * Retrieves only OPEN ASNs.... for Perf test generation
 */
AsnHelper.prototype.getOpenAsns = function() {
    var self = this;
    var where = this.getOpenWhere();
    var out = {
        count: 0,
        rows: null
    };
    return new Promise(function (resolve, reject) {
        self.models.ASN.count({where: where}).then(function (c) {
            out.count = c;
            self.models.ASN.findAll({
                where: where,
                include: [{model: self.models.AsnAddress, as: 'Addresses'}]
            }).then(function (data) {
                out.rows = data;
                resolve(out);
            }).catch(function (err) {
                reject(err);
            });
        }).catch(function (cErr) {
            reject(cErr);
        });
    });
};

/**
 * Common routine for querying ASNs for the Web API
 *
 * @param {Object} req ExpressJS Request
 */
AsnHelper.prototype.getAllASNs = function(req) {
    var self = this;
    var limit = getLimit(req);
    var offset = getOffset(req);
    var order = getOrder(req);
    var where = getWhere(req);

    if (limit) limit = Number(limit);
    if (offset) offset = Number(offset);
    if (order) {
        order = order.split(" ");
        if (order[0].indexOf("'") >=0) order[0] = order[0].replace(/'/g, "");
        order = [order];
    }
    if (where) {
        //a where clause alters counts as well
        try {
            var whereArgs = self.buildWhereArgs(where, 'ASN', self.models.ASN.rawAttributes);
            return new Promise(function (resolve, reject) {
                self.models.ASN.findAndCountAll({
                    where: whereArgs,
                    limit: limit,
                    offset: offset,
                    order: order
                }).then(function (data) {
                    resolve(data);
                }).catch(function (err) {
                    reject(err);
                });
            });
        } catch(err) {
            console.error(err);
        }
    } else {
        //no where clause = pull ALL for Web, paginated
        return new Promise(function (resolve, reject) {
            self.models.ASN.findAndCountAll({
                limit: limit,
                offset: offset,
                order: order
            }).then(function (data) {
                resolve(data);
            }).catch(function (cErr) {
                reject(cErr);
            });
        });
    }
};

/**
 * Common routine for querying ASNs (with Addresses) for the Web API
 *
 * NOTE: Sequelize BUG will not allow an Order By clause
 *
 * @param {Object} req ExpressJS Request
 */
AsnHelper.prototype.getAllASNsWithAddresses = function(req) {
    var self = this;
    var limit = getLimit(req);
    var offset = getOffset(req);
    var where = getWhere(req);
    var out = {
        count: 0,
        rows: null
    };
    if (limit) limit = Number(limit);
    if (offset) offset = Number(offset);
    if (where) {
        //a where clause alters counts as well
        var whereArgs = self.buildWhereArgs(where, 'ASN', self.models.ASN.rawAttributes);
        return new Promise(function (resolve, reject) {
            self.models.ASN.count({ where: whereArgs }).then(function (c) {
                out.count = c;
                self.models.ASN.findAll({
                    where: whereArgs,
                    include: [{model: self.models.AsnAddress, as: 'Addresses'}],
                    limit: limit,
                    offset: offset
                }).then(function (data) {
                    out.rows = data;
                    resolve(out);
                }).catch(function (err) {
                    reject(err);
                });
            }).catch(function (cErr) {
                reject(cErr);
            });
        });
    } else {
        //no where clause = pull ALL for Web, paginated
        return new Promise(function (resolve, reject) {
            self.models.ASN.count().then(function (c) {
                out.count = c;
                self.models.ASN.findAll({
                    include: [{model: self.models.AsnAddress, as: 'Addresses'}],
                    limit: limit,
                    offset: offset
                }).then(function (data) {
                    out.rows = data;
                    resolve(out);
                }).catch(function (err) {
                    reject(err);
                });
            }).catch(function (cErr) {
                reject(cErr);
            });
        });
    }
};

/**
 * Common routine for pulling the Order By clause from the req object
 * @param {Object} req  ExpressJS Request
 * @returns {Object} null or Order By JSON
 */
function getOrder(req) {
    var def = null;
    var order = def;
    if (req.body && req.body.order) {
        order = req.body.order;
    } else if (req.query && req.query.order) {
        order = req.query.order;
    }
    return order;
}
/**
 * Common routine for pulling the Limit clause from the req object
 * @param {Object} req  ExpressJS Request
 * @returns {String} null or limit
 */
function getLimit(req) {
    var limit = null;
    if (req.body && req.body.limit) {
        limit = req.body.limit;
    } else if (req.query && req.query.limit) {
        limit = req.query.limit;
    }

    return limit;
}
/**
 * Common routine for pulling the Offset clause from the req object
 * @param {Object} req  ExpressJS Request
 * @returns {String} null or offset
 */
function getOffset(req) {
    var offset = 0;
    if (req.body && req.body.offset) {
        offset = req.body.offset;
    } else if (req.query && req.query.offset) {
        offset = req.query.offset;
    }
    return offset;
}
/**
 * Common routine for pulling the Where clause from the req object
 * @param {Object} req  ExpressJS Request
 * @returns {Object} null or where JSON
 */
function getWhere(req) {
    var where = null;
    if (req.body && req.body.where) {
        where = req.body.where;
    } else if (req.query && req.query.where) {
        where = req.query.where;
    }
    return where;
}
/**
 * Devices will be requesting their own ASNs.
 *
 * This centralizes the logic for determining which ASNs are eligible to be pulled.
 *
 * @param {Object} req  ExpressJS Request
 * @returns {Promise.<Array.<Model>>|*} The findAll promise (to be executed and handled by caller)
 */
AsnHelper.prototype.getAllASNsForClinic = function(req) {
    var self = this;
    var limit = getLimit(req);
    var offset = getOffset(req);
    var order = getOrder(req);
    var where = this.getOpenWhere();
    where['clinicId'] = req.params.clinic;

   return this.models.ASN.findAll({
        where: where,
        limit: limit,
        offset: offset,
        order: order,
        include: [{model: self.models.AsnAddress, as: 'Addresses'}]
    });
};
/**
 * Common definition for "Open" ASNs
 *
 * @returns {Sequelize.where} Sequelize where argument object
 */
AsnHelper.prototype.getOpenWhere = function() {
    var self = this;
    return {
        lostQty: 0,
        receivedQty: {
            $lt: self.models.Sequelize.col(self.models.ASN.rawAttributes['shipQty'].field)
        }

    };
};


/**
 * Converts the Where clause JSON into sequelize JSON
 *
 * @param {Object} where clause from ExpressJS request
 * @param {String} objName The Sequelize Model name
 * @param {Object} objRawAttributes The Sequelize Model's raw attributes (mapping object)
 * @returns {Array} Collection of Sequelize where clauses
 */
AsnHelper.prototype.buildWhereArgs = function(where, objName, objRawAttributes) {
    var self = this;
    var whereArgs = [];
    if (where != null) {
        this.logger.debug(JSON.stringify(where));
        for (var key in where) {
            if (objRawAttributes.hasOwnProperty(key)) {
                var newCondition = String(where[key]);
                if (newCondition === "[object Object]") {
                    if (key === "timestamp" || key === 'asnDate') {
                        //we have gt/lt range
                        if (where[key]['$gt']) {
                            var newWhere = {};
                            newWhere[objRawAttributes[key].field] = {$gt: moment(where[key]['$gt'])};
                            whereArgs.push(newWhere);
                        }
                        if (where[key]['$lt']) {
                            var newWhere = {};
                            newWhere[objRawAttributes[key].field] = {$lt: moment(where[key]['$lt'])};
                            whereArgs.push(newWhere);
                        }
                        continue;
                    } else {
                        newCondition = String(where[key]['$like']);
                    }
                }
                var dataType = String(objRawAttributes[key].type);
                if (newCondition.indexOf("*") < 0) {
                    newCondition = "%" + newCondition + "%";
                } else {
                    newCondition = newCondition.replace(/\*/g, "%");
                }
                if (dataType.indexOf("VARCHAR") < 0) {
                    //cast numerics as strings so we can use the "LIKE" operator
                    if (dataType.indexOf("TIME") > 0) {
                        //since time is hidden in the UI, we should strip time from the date fields
                        whereArgs.push(self.models.Sequelize.where(self.models.Sequelize.fn('DATE_FORMAT', self.models.Sequelize.col(objName + "." + objRawAttributes[key].field) ,'%m-%d-%Y'), 'LIKE', newCondition));
                    } else {
                        //var newWhere = self.models.Sequelize.where(self.models.Sequelize.cast(self.models.Sequelize.col(objName + "." + objRawAttributes[key].field), 'CHAR(255)'), 'LIKE', newCondition);
                        var newWhere = {};

                        newWhere[objRawAttributes[key].field] = {$like: newCondition};
                        whereArgs.push(newWhere);
                    }
                    delete where[key];
                } else {
                    //already a String... just use the LIKE operator
                    var newWhere = {};

                    newWhere[objRawAttributes[key].field] = {$like: newCondition};
                    whereArgs.push(newWhere);
                }
            }
        };
    }
    return whereArgs;
};


module.exports = AsnHelper;
