"use strict";

var fs          = require("fs");
var path        = require("path");
var Sequelize   = require("sequelize");
var env         = process.env.NODE_ENV || "dev";
var config      = require(path.join(__dirname,  '..', '/config/server.' + env + '.js'))['db'];
config.logging = console.log;
config.omitNull = true;
var sequelize = new Sequelize(config.database, config.username, config.password, config);
var db        = {};

fs
    .readdirSync(__dirname)
    .filter(function(file) {
        return (file.indexOf(".") !== 0) && (file !== "index.js");
    })
    .forEach(function(file) {
        var model = sequelize.import(path.join(__dirname, file));
        db[model.name] = model;
    });

Object.keys(db).forEach(function(modelName) {
    if ("associate" in db[modelName]) {
        db[modelName].associate(db);
    }
});

module.exports = function setupReduceAll(Sequelize) {
    'use strict';

    // findAll in chunks and reduce to single value.
    // This is done in a memory efficient way by taking x rows each time.
    // Where x is defined by options.limit.
    //
    // options :      options passed to findAll. options.limit defines how many rows
    //                should be reduced per iteration.
    // action :       callback that is called for each iteration.
    //                gets parameter data and previous reduce value.
    //                The callback should return new reduce value
    //                or a Promise that returns a reduced value.
    // initialValue : Initial value of for the reduction.
    //
    // Returns Promise with reduced value
    //
    // Example:
    //    var initialSum = 0;
    //    Model
    //        .reduceAll({
    //            limit: 100
    //        }, function (data, sum) {
    //            return data.reduce(function (row, currentSum) {
    //                return currentSum + row.number;
    //            }, sum);
    //        }, initialSum)
    //        .then(function (totalSum) {
    //            console.info('Total sum is: ', totalSum);
    //        });

    Sequelize.Model.prototype.reduceAll = function (options, action, initialValue) {
        var self = this,
            foundRecords = -1,
            reducedValue = initialValue;

        options = options || {};
        options.limit = options.limit || 1000;
        options.offset = 0;

        function reduceNextChunk() {
            if (foundRecords === 0) {
                // No more records found in table
                // return the final reduced value.
                return reducedValue;
            } else {
                return self
                    .findAll(options)
                    .tap(function (data) {
                        foundRecords = data.length;
                        // Increase the offset for next interation.
                        options.offset += options.limit;
                    })
                    .then(function (data) {
                        // Call the callback with data rows and last reduced value.
                        return action(data, reducedValue);
                    })
                    .then(function (newReducedValue) {
                        // Update reduced value
                        reducedValue = newReducedValue;
                    })
                    // Find next chunk
                    .then(reduceNextChunk);
            }
        }

        return reduceNextChunk();
    };
};

//Sequelize.INTEGER.prototype.validate = function() { return true; };


db.sequelize = sequelize;
db.Sequelize = Sequelize;

module.exports = db;

